import time
# import math
# from typing import Tuple

_start = time.time()


def get_timer():
    end = time.time()
    return end - _start


def lerp(a, b, d):
    return a * (1 - d) + b * d
# print lerp([3, 5], 0.75)


# def distance(point1, point2):
#     """ Return direct distance from point1 to point2. """
#     x_dist = abs(point1[0] - point2[0])
#     y_dist = abs(point1[1] - point2[1])
#     dist = math.sqrt(x_dist**2 + y_dist**2)
#     return dist



"""
    domonic.geom
    ====================================

    much of this was written by.ai
    (https://6b.eleuther.ai/)

    most of this code was derivied by continually pasting back into eleuther

    # TODO - tests for all these new functions. sadly the ai didn't generate those.

"""
import math
from domonic.javascript import Math


class vec2():
    """[vec2]
    """

    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def add(self, pt):
        """ Adds a point to this point. returns new one """
        return vec2(self.x + pt.x, self.y + pt.y)

    def __add__(self, other):
        self.add(other)

    def subtract(self, pt):
        """ Subtract from this point. returns new one """
        return vec2(self.x - pt.x, self.y - pt.y)

    def __sub__(self, other):
        if isinstance(other, vec2):
            return self.x - other.x and self.y - other.y
        else:
            return self.x - other and self.y - other

    def __eq__(self, other):
        if isinstance(other, vec2):
            return self.x == other.x and self.y == other.y
        else:
            return self.x == other and self.y == other

    def __ne__(self, other):
        if isinstance(other, vec2):
            return self.x != other.x or self.y != other.y
        else:
            return self.x != other and self.y != other

    def __gt__(self, other):
        if isinstance(other, vec2):
            return self.x > other.x or self.y > other.y
        else:
            return self.x > other and self.y > other

    def __ge__(self, other):
        if isinstance(other, vec2):
            return self.x >= other.x or self.y >= other.y
        else:
            return self.x >= other and self.y >= other

    def __lt__(self, other):
        if isinstance(other, vec2):
            return self.x < other.x or self.y < other.y
        else:
            return self.x < other and self.y < other

    def __le__(self, other):
        if isinstance(other, vec2):
            return self.x <= other.x or self.y <= other.y
        else:
            return self.x <= other and self.y <= other

    def __mul__(self, other):
        if isinstance(other, vec2):
            return self.__class__((self.x * other.x, self.y * other.y))
        else:
            return self.__class__((self.x * other, self.y * other))

    def __truediv__(self, other):
        if isinstance(other, vec2):
            return self.__class__((self.x / other.x, self.y / other.y))
        else:
            return self.__class__((self.x / other, self.y / other))

    def __div__(self, other):
        if isinstance(other, vec2):
            return self.__class__((self.x / other.x, self.y / other.y))
        else:
            return self.__class__((self.x / other, self.y / other))

    def __mod__(self, other):
        if isinstance(other, vec2):
            return self.__class__(self.x % other.x, self.y % other.y)
        else:
            return self.__class__(self.x % other, self.y % other)

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.x
            elif item == 1:
                return self.y
        elif isinstance(item, str):
            if item == 'x':
                return self.x
            elif item == 'y':
                return self.y

    def copy(self):
        """ Creates a copy of this Point object. """
        return vec2(self.x, self.y)

    def dot(self, other):
        """ Returns the dot product of these two vectors """
        x1 = self.x
        y1 = self.y
        x2 = other.x
        y2 = other.y
        return x1 * x2 + y1 * y2

    def cross(self, other):
        """ Returns the cross product of this and other. The cross product is considered 0
        perpendicular to each other point """
        x1 = self.x
        y1 = self.y
        x2 = other.x
        y2 = other.y
        return x1 * y2 - y1 * x2

    def negative(self):
        """ return vector but negative version """
        return vec2(-self.x, -self.y)

    def distance(self, other):
        x_dot = self.x - other.x
        y_dot = self.y - other.y
        return math.sqrt(x_dot * x_dot + y_dot * y_dot)

    def squareDistance(self, other):
        return self.distance(self) * self.distance(other)

    def equals(self, other):
        """ Returns true if all vectors components are the same """
        return (self.x == other.x and self.y == other.y)

    def interpolate(self, pt1, pt2, t):
        """ Calculates the point that would be reached by this Point
        if we moved by a given distance vector over time t.
        Only floats are handled here. """
        return vec2(self.x * (1 - t) + pt1.x * t, self.y * (1 - t) + pt1.y * t)

    def rotateAround(self, target, angle):
        """Rotates the vector around another point. In fact it returns the other point."""
        dot = self.dot(target)
        s = math.sin(angle / 2)
        c = math.cos(angle / 2)
        x = s * (self.x - dot * c) + c
        y = s * (self.y - dot * c) - c
        return vec2(x, y)

    def copyRotateAround(self, target, angle):
        """moves the vector then rotate"""
        self.rotateAround(target, angle)
        return vec2(self.x, self.y)

    def mirrorHorizontally(self):
        return self.rotateAround(vec2(1, 0), 90)

    def mirrorVertically(self):
        return self.rotateAround(vec2(0, 1), 90)

    def negate(self):
        """Negate a vector. This results in vector with the same direction but different length"""
        return vec2(-self.x, -self.y)

    def length(self):
        """ returns the length of this vector """
        return self.x

    def squaredlength(self):
        """ returns the squared length of this vector """
        return self.x * self.x + self.y * self.y

    def __str__(self):
        return str(self.x) + " " + str(self.y)


class vec3():
    """[vec3]
    """

    def __init__(self, x=0, y=0, z=0, w=0):
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def __add__(self, other):
        if isinstance(other, vec3):
            return self.__class__((self.x + other.x, self.y + other.y, self.z + other.z, self.w + other.w))
        else:
            return self.__class__((self.x + other, self.y + other, self.z + other, self.w + other))

    def __sub__(self, other):
        if isinstance(other, vec3):
            return self.__class__((self.x - other.x, self.y - other.y, self.z - other.z, self.w - other.w))
        else:
            return self.__class__((self.x - other, self.y - other, self.z - other, self.w - other))

    def __mul__(self, other):
        if isinstance(other, vec3):
            return self.__class__((self.x * other.x, self.y * other.y, self.z * other.z, self.w * other.w))
        else:
            return self.__class__((self.x * other, self.y * other, self.z * other, self.w * other))

    # def __rmul__(self, other):
    #     return vec3(other.x * self.x, other.y * self.y, other.z * self.z, other.w * self.w)

    def __truediv__(self, other):
        if isinstance(other, vec3):
            return self.__class__(self.x / other.x, self.y / other.y, self.z / other.z, self.w / other.w)
        else:
            return self.__class__(self.x / other, self.y / other, self.z / other, self.w / other)

    # def __pow__(self, other):
    #     return vec3(self.x ** other.x, self.y ** other.y, self.z ** other.z, self.w ** other.w)

    # def __mod__(self, other):
    #     return vec3(self.x % other.x, self.y % other.y, self.z % other.z, self.w % other.w)

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.x
            elif item == 1:
                return self.y
            elif item == 2:
                return self.z
            elif item == 3:
                return self.w
        elif isinstance(item, str):
            if item == 'x':
                return self.x
            elif item == 'y':
                return self.y
            elif item == 'z':
                return self.z
            elif item == 'w':
                return self.w

    def __iadd__(self, other):
        self.x += other.x
        self.y += other.y
        self.z += other.z
        self.w += other.w

    def add(self, point):
        self.x += point.x
        self.y += point.y
        self.z += point.z
        return self

    def subtract(self, point):
        """ Subtract from this point. """
        self.x -= point.x
        self.y -= point.y
        self.z -= point.z
        return self

    def dot(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z + self.w * other.w

    def cross(self, other):
        return self.x * other.y - self.y * other.x, self.y * other.z - self.z * other.y, self.z * other.x - self.x * other.z

    def mul(self, v):
        return (v.x * self.x + v.y * self.y + v.z * self.z + v.w * self.w)

    def copy(self):
        """ Creates a copy of this object. """
        return vec3(self.x, self.y, self.z, self.w)

    def angleBetween(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        dz = self.z - other.z
        return math.sqrt(dx * dx + dy * dy + dz * dz)

    def length(self):
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w)

    def distance(self, other):
        """ Returns the distance between this point and another vector3."""
        return (self.x - other.x)**2 + (self.y - other.y)**2 + (self.z - other.z)**2 + (self.w - other.w)**2
        # return sqrt(pow(self.x - other.x, 2) + pow(self.y - other.y, 2) + pow(self.z - other.z, 2) + pow(self.w - other.w, 2))

    def equals(self, other):
        """ Determine whether two objects are identical. """
        return self.x == other.x and self.y == other.y and self.z == other.z and self.w == other.w

    def intersects(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        dz = self.z - other.z
        dw = self.w - other.w
        return (dx * dy - dy * dz) > 0 and (dx * dw - dw * dz) > 0 and (dx * dz + dz * dw) > 0 and (dy * dw - dw * dy) > 0 and (dy * dw - dw * dw) > 0

    def clone(self):
        """ Returns a new instance of this vector3. """
        return vec3(self.x, self.y, self.z, self.w)

    def apply(self, point, amount):
        """ Moves the points x,y,z,w by amount. """
        return vec3(point.x + amount.x, point.y + amount.y, point.z + amount.z, point.w + amount.w)

    def __str__(self):
        return str(self.x) + " " + str(self.y) + " " + str(self.z)


class vec4():
    """[vec2]
    """

    def __init__(self, x=0, y=0, z=0, w=0, n=1):
        self.x = x
        self.y = y
        self.z = z
        self.w = w
        self.n = n

    def __add__(self, other):
        x = self.x + other.x
        y = self.y + other.y
        z = self.z + other.z
        w = self.w + other.w
        return vec4(x, y, z, w)

    def __sub__(self, other):
        x = self.x - other.x
        y = self.y - other.y
        z = self.z - other.z
        w = self.w - other.w
        return vec4(x, y, z, w)

    def __mul__(self, other):
        n = self.n * other.n
        a = self.x * other.x + self.y * other.y + self.z * other.z
        b = self.x * other.y - self.y * other.x + self.z * other.w
        c = self.x * other.z - self.z * other.x + self.y * other.w
        d = self.w * other.n - self.x * other.w
        x = a + b * n
        y = b - a * n
        z = c + d * n
        w = d - c * n
        return vec4(x, y, z, w)


class matrix(object):
    """[matrixs]
    """

    def __init__(self, m):
        self.m = m

    def translate(self, pt):
        """ Translates the point on the vector defined by the vectors m[0] and m[1]."""
        return vec2(self.m[0][0], self.m[0][1]) + pt - self.m[0][0]


# class Quaternion():

class vertex(object):
    
    def __init__(self, x=0, y=0):
        self.x = x
        self.y = y

    def move_to(self, position):
        x, y = position
        self.x = x
        self.y = y


Point = vertex


class Shape(Point):

    def __init__(self, x=0, y=0, color='red'):
        super().__init__(x, y)
        self.color = color
        # self.id
        # self.name
        # self.style
        # self.locked
        # self.visible
        # self.blendMode
        # self.opacity
        # self.selected
        # self.clipMask
        # self.data

        # self.strokeColor
        # self.strokeWidth
        # self.strokeCap
        # self.strokeJoin
        # self.dashOffset
        # self.strokeScaling
        # self.dashArray
        # self.miterLimit

        # self.fillColor

        # self.shadowColor
        # self.shadowBlur
        # self.shadowOffset

    # def scale_to(self, scale):
        # self.size *= scale

    # def draw(self):
    # def clear(self):

    @property
    def position(self):
        return Point(self.x, self.y)

    # def pivot
    # def bounds
    # def strokeBounds
    # def handleBounds
    # def internalBounds
    # def rotation
    # def scaling
    # def matrix
    # def globalMatrix
    # def viewMatrix
    # def applyMatrix

    # def rasterize()

    # def contains(point)
    # def isInside(rect)
    # def intersects(item)


class Line(Shape):
    def __init__(self, p1, p2, color=None):
        super().__init__(color)
        self.p1 = p1
        self.p2 = p2
        self.x = self.p1[0] - self.p2[0]
        self.y = self.p1[1] - self.p2[1]


class Rect(Shape):
    def __init__(self, x=0, y=0, width=1, height=1, color=None):
        super().__init__(color)
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def get_bottom_right(self):
        d = self.x + self.width
        t = self.y + self.height
        return (d, t)

    def get_bottom_left(self):
        b = self.get_bottom_right()
        l = super().get_bottom_right()
        return (b[0], l[0])

    def get_top_left(self):
        t = self.x
        b = super().get_bottom_right()
        return (t, b[1])


class Square(Rect):
    def __init__(self, x=0, y=0, size=1.0, color=None):
        super().__init__(x, y, size, size, color)


class Polygon(Shape):
    def __init__(self, points):
        super().__init__()
        self.points = points


class Polyline(Shape):
    def __init__(self, points):
        super().__init__()
        self.points = points


# class Quad(Shape):
#     def __init__(self, left=Rect(), right=Rect(), front=Rect(), back=Rect()):
#         super().__init__(color='blue')
#         self.left, self.right, self.front, self.back = left, right, front, back

# class Triangle(Shape):
#     def __init__(self):
#         super().__init__(self.color='blue')


class Circle(Shape):

    def __init__(self, x, y, radius=1.0, color=None):
        super().__init__(color)
        self.radius = radius  # Create an instance variable radius

    @property
    def area(self):
        return self.radius * self.radius * Math.PI

    @property
    def perimeter(self):
        return 2 * self.radius * Math.PI

    @property
    def average_circumference(self):
        return 2 * self.radius

    @property
    def center(self):
        x, y = self.center = [self.radius, self.radius]
        return x, y


class Oval(Circle):
    def __init__(self, radius=2.5, size=3):
        super().__init__(size, 'green')
        self.radius = radius
        self.x = self.width / 2 - self.radius / 2
        self.y = self.width / 2 + self.radius / 2


class Ellipse(Shape):
    def __init__(self, x, y, width, height, color=None):
        super().__init__(color)
        self.x = x
        self.y = y
        self.width = width
        self.height = height


class Plotter(object):
    def __init__(self, canvas=None):
        self.canvas = canvas  # or canvas.Canvas('Example Plotter', 500, 500)
        self.points = []
        self.starting_color = (255, 255, 255)
        self.color = self.starting_color
        self.clear()

    def add_point(self, point):
        self.points.append(point)

    # def clear(self):
    #     pass


# class Path(object):
# class Cursor(object):
# class Group(object):
# class Layer(object):


# class Interactive(object):#??
    # def handle_click_event(self):

    # onFrame
    # onMouseDown
    # onMouseDrag
    # onMouseUp
    # onClick
    # onDoubleClick
    # onMouseMove
    # onMouseEnter
    # onMouseLeave



"""
    domonic.utils
    ====================================
    snippets etc
"""
import typing
import random
from re import sub
from itertools import chain, islice
from collections import Counter


class Utils(object):
    """ utils """

    @staticmethod
    def case_camel(s: str):
        """ case_camel('camel-case') > 'camelCase' """
        s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
        return s[0].lower() + s[1:]

    @staticmethod
    def case_snake(s: str):
        """
        snake('camelCase') # 'camel_case'
        """
        return '_'.join(
            sub('([A-Z][a-z]+)', r' \1',
            sub('([A-Z]+)', r' \1',
            s.replace('-', ' '))).split()).lower()

    @staticmethod
    def case_kebab(s: str):
        """
        kebab('camelCase') # 'camel-case'
        """
        return '-'.join(
            sub(r"(\s|_|-)+", " ",
            sub(r"[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+",
            lambda mo: ' ' + mo.group(0).lower(), s)).split())

    @staticmethod
    def squash(the_list: list):
        """[turns a 2d array into a flat one]

        Args:
            the_list ([type]): [a 2d array]

        Returns:
            [type]: [a flattened 1d array]
        """
        return [inner for outer in the_list for inner in outer]

    @staticmethod
    def chunk(list: list, size: int):
        """ chunk a list into batches """
        return [list[i:i + size] for i in range(0, len(list), size)]

    @staticmethod
    def dictify(arr: str):
        """[turns a list into a dictionary where the list items are the keys]

        Args:
            arr ([type]): [list to change]

        Returns:
            [type]: [a new dict where the list items are now the keys]
        """
        return dict().fromkeys(arr, 0)

    @staticmethod
    def is_empty(some_str):
        return (not some_str.strip())

    @staticmethod
    def unique(some_arr):
        """[removes duplicates from a list]

        Args:
            some_arr ([type]): [list containing duplicates]

        Returns:
            [type]: [a list containing no duplicates]
        """
        return list(set(some_arr))

    @staticmethod
    def chunks(iterable, size, format=iter):
        """ Iterate over any iterable (list, set, file, stream, strings, whatever), of ANY size """
        it = iter(iterable)
        while True:
            yield format(chain((it.next(),), islice(it, size - 1)))
    # >>> l = ["a", "b", "c", "d", "e", "f", "g"]
    # >>> for chunk in chunks(l, 3, tuple):
    # ...         print chunk

    @staticmethod
    def clean(lst):
        """[removes falsy values (False, None, 0 and “”) from a list ]

        Args:
            lst ([type]): [lst to operate on]

        Returns:
            [type]: [a new list with falsy values removed]
        """
        return list(filter(None, lst))

    @staticmethod
    def get_vowels(string: str):
        """[get a list of vowels from the word]

        Args:
            string ([str]): [the word to check]

        Returns:
            [list]: [a list of vowels]
        """
        return [each for each in string if each in 'aeiou']

    @staticmethod
    def untitle(string: str):
        """[the opposite of title]

        Args:
            str ([type]): [the string to change]

        Returns:
            [type]: [a string with the first character set to lowercase]
        """
        return string[:1].lower() + string[1:]

    @staticmethod
    def merge_dictionaries(a, b):
        """[merges 2 dicts]

        Args:
            a ([type]): [dict a]
            b ([type]): [dict b]

        Returns:
            [dict]: [a new dict]
        """
        return {**a, **b}

    @staticmethod
    def to_dictionary(keys, values):
        """[take a list of keys and values and returns a dict]

        Args:
            keys ([type]): [a list of keys]
            values ([type]): [a list of value]

        Returns:
            [dict]: [a dictionary]
        """
        return dict(zip(keys, values))

    @staticmethod
    def most_frequent(list):
        return max(set(list), key=list.count)

    @staticmethod
    def anagram(first, second):
        return Counter(first) == Counter(second)

    @staticmethod
    def acronym(sentence: str):
        """[pass a sentence, returns the acronym]

        Args:
            sentence ([str]): [typically 3 words]

        Returns:
            [str]: [a TLA (three letter acronym)]
        """
        text = sentence.split()
        a = ""
        for i in text:
            a = a + str(i[0]).upper()
        return a

    @staticmethod
    def frequency(data):
        """[check the frequency of elements in the data]

        Args:
            data ([type]): [the data to check]

        Returns:
            [dict]: [a dict of elements and their frequency]
        """
        freq = {}
        for elem in data:
            if elem in freq:
                freq[elem] += 1
            else:
                freq[elem] = 1
        return freq

    @staticmethod
    def init_assets(dir: str = 'assets'):
        """[creates an assets directory with nested js/css/img dirs]

        Args:
            dir (str, optional): [default directory name]. Defaults to 'assets'.
        """
        from domonic.terminal import mkdir, touch
        mkdir(f"{dir}")
        mkdir(f"{dir}/js")
        mkdir(f"{dir}/css")
        mkdir(f"{dir}/img")
        touch(f"{dir}/js/master.js")
        touch(f"{dir}/css/style.css")
        return

    @staticmethod
    def url2file(url: str):
        """[gen a safe filename from a url. by replacing '/' for '_' and ':' for '__' ]

        Args:
            url ([str]): [the url to turn into a filename]

        Returns:
            [str]: [description]
        """
        import urllib
        url = "_".join(url.split("/"))
        url = "__".join(url.split(":"))
        filename = urllib.parse.quote_plus(url, '')
        return filename

    @staticmethod
    def permutations(word: str):
        """[provides all the possible permutations of a given word]

        Args:
            word ([str]): [the word to get permutations for]

        Returns:
            [str]: [description]
        """
        from itertools import permutations
        return [''.join(perm) for perm in list(permutations(word))]

    @staticmethod
    def random_color(self):
        ''' TODO - remove in 0.3 as we have color class. '''
        r = lambda: random.randint(0, 255)
        return str('#%02X%02X%02X' % (r(), r(), r()))

    @staticmethod
    def escape(s: str):
        """[escape a string]

        Args:
            s ([str]): [the string to escape]

        Returns:
            [str]: [description]
        """
        chars = {
            "&": "&amp;",
            '"': "&quot;",
            "'": "&apos;",
            ">": "&gt;",
            "<": "&lt;"
        }
        return "".join(chars.get(c, c) for c in s)

    @staticmethod
    def unescape(s: str):
        """[unescape a string]

        Args:
            s ([str]): [the string to unescape]

        Returns:
            [str]: [description]
        """
        s = s.replace("&lt;", "<")
        s = s.replace("&gt;", ">")
        s = s.replace("&quot;", '"')
        s = s.replace("&apos;", "'")
        s = s.replace("&amp;", "&")
        return s

    @staticmethod
    def replace_between(content: str, match: str, replacement: str, start: int = 0, end: int = 0):
        """[replace some text but only between certain indexes]

        Args:
            content (str): [the content whos text you will be replacing]
            match (str): [the string to find]
            replacement (str): [the string to replace it with]
            start (int, optional): [start index]. Defaults to 0.
            end (int, optional): [end index]. Defaults to 0.

        Returns:
            [str]: [description]
        """
        front = content[0:start]
        mid = content[start:end]
        end = content[end:len(content)]
        mid = mid.replace(match, replacement)
        return front + mid + end

    @staticmethod
    def truncate(text='', length=0):
        """[truncates a string and appends 3 dots]

        Args:
            text (str, optional): [the text to truncate]. Defaults to ''.
            length (int, optional): [the max length]. Defaults to 0.

        Returns:
            [str]: [description]
        """
        if len(text) > length:
            return text[0:length] + "..."
        else:
            return text + "..."

    @staticmethod
    def has_internet(url='http://www.google.com/', timeout=5):
        import requests
        try:
            _ = requests.head(url, timeout=timeout)
            return True
        except requests.ConnectionError:
            print("No internet connection available.")
        return False

    # def any_iter(arr):
    #     ''' given a list. returns random until expired '''
    #     random.shuffle(arr)
    #     return (x for x in arr)

    # @staticmethod
    # def unless(value, condition):
        # return value if condition else not value
        # if any(pred(x.item) for x in sequence):

    # TODO -
    # def beautfiy(): # make nice
    # def uglify(): # make not nice
    # def simplify(sentence): # reduce a sentence to its meaning. remove uneeded words.
    # def factualise():  # returns json document of modelled info from general text



"""
    domonic.particles
    ====================================
    fun stuff

"""

from domonic.javascript import *


class Particle():

    def __init__(self, size=1):
        self.x = 0
        self.y = 0
        self.width = size
        self.height = size
        self.rotation = 0
        self.vx = 0
        self.vy = 0
        self.damp = 0.9
        self.bounce = -0.5
        self.grav = 0
        self.wander = 0
        self._k = 0.2
        self.edges = "wrap"
        self._drag = False
        self._oldx = None
        self._oldy = None
        self._turnToPath = False
        self._bounds = {}
        self.set_bounds({'xMin': 0, 'yMin': 0, 'yMax': 800, 'xMax': 800})
        self.maxSpeed = Number.MAX_VALUE

    def set_bounds(self, oBounds):
        self._bounds['top'] = oBounds['yMin']
        self._bounds['bottom'] = oBounds['yMax']
        self._bounds['left'] = oBounds['xMin']
        self._bounds['right'] = oBounds['xMax']

    def turnToPath(self, bTurn):
        self._turnToPath = bTurn

    def update(self):
        if self._drag:
            self.vx = self.x - self._oldx
            self.vy = self.y - self._oldy
            self._oldx = self.x
            self._oldy = self.y

        else:
            self.vx += Math.random() * self.wander - self.wander / 2
            self.vy += Math.random() * self.wander - self.wander / 2
            self.vy += self.grav
            self.vx *= self.damp
            self.vy *= self.damp

            speed = Math.sqrt(self.vx * self.vx + self.vy * self.vy)
            if speed > self.maxSpeed:
                self.vx = self.maxSpeed * self.vx / speed
                self.vy = self.maxSpeed * self.vy / speed

            if self._turnToPath:
                self.rotation = Math.atan2(self.vy, self.vx)

            self.x += self.vx
            self.y += self.vy
            if self.edges == "wrap":
                if self.x > (self._bounds['right'] + self.width / 2):
                    self.x = self._bounds['left'] - self.width / 2
                elif self.x < self._bounds['left'] - self.width / 2:
                    self.x = self._bounds['right'] + self.width / 2

                if(self.y > self._bounds['bottom'] + self.height / 2):
                    self.y = self._bounds['top'] - self.height / 2
                elif (self.y < self._bounds['top'] - self.height / 2):
                    self.y = self._bounds['bottom'] + self.height / 2

            elif self.edges == "bounce":
                if self.x > (self._bounds['right'] - self.width / 2):
                    self.x = self._bounds['right'] - self.width / 2
                    self.vx *= self.bounce
                elif self.x < (self._bounds['left'] + self.width / 2):
                    self.x = self._bounds['left'] + self.width / 2
                    self.vx *= self.bounce

                if self.y > (self._bounds['bottom'] - self.height / 2):
                    self.y = self._bounds['bottom'] - self.height / 2
                    self.vy *= self.bounce
                elif self.y < (self._bounds['top'] + self.height / 2):
                    self.y = self._bounds['top'] + self.height / 2
                    self.vy *= self.bounce

            # else if self.edges == "remove":
            #     if(self.x > self._bounds.right + self.width/2 || self.x<self._bounds.left - self.width/2 ||
            #        self.y > self._bounds.bottom + self.height/2 || self.y<self._bounds.top - self.height/2){

        if Global.isNaN(self.x):
            self.x = self.y = self.vx = self.vy = 0


class Particle3D():

    def __init__(self, size=1):
        self.x = 0
        self.y = 0
        self.z = 0
        self.width = size
        self.height = size
        self.length = size
        self.rotation = 0
        self.vx = 0
        self.vy = 0
        self.vz = 0
        self.damp = 0.9
        self.bounce = -0.5
        self.grav = 0
        self.wander = 0
        self._k = 0.2
        self.edges = "wrap"
        self._drag = False
        self._oldx = None
        self._oldy = None
        self._oldz = None
        self._turnToPath = False
        self._bounds = {}
        self.set_bounds({'xMin': 0, 'yMin': 0, 'yMax': 800, 'xMax': 800})
        self.maxSpeed = Number.MAX_VALUE

    def set_bounds(self, oBounds):
        self._bounds['top'] = oBounds['yMin']
        self._bounds['bottom'] = oBounds['yMax']
        self._bounds['left'] = oBounds['xMin']
        self._bounds['right'] = oBounds['xMax']

    def turnToPath(self, bTurn):
        self._turnToPath = bTurn

    def update(self):
        if self._drag:
            self.vx = self.x - self._oldx
            self.vy = self.y - self._oldy
            self.vz = self.z - self._oldz
            self._oldx = self.x
            self._oldy = self.y
            self._oldz = self.z

        else:
            self.vx += Math.random() * self.wander - self.wander / 2
            self.vy += Math.random() * self.wander - self.wander / 2
            self.vz += Math.random() * self.wander - self.wander / 2
            self.vy += self.grav
            self.vx *= self.damp
            self.vy *= self.damp
            self.vz *= self.damp

            speed = Math.sqrt(self.vx * self.vx + self.vy * self.vy + self.vz * self.vz)
            if speed > self.maxSpeed:
                self.vx = self.maxSpeed * self.vx / speed
                self.vy = self.maxSpeed * self.vy / speed
                self.vz = self.maxSpeed * self.vz / speed

            if self._turnToPath:
                self.rotation = Math.atan2(self.vy, self.vx)

            self.x += self.vx
            self.y += self.vy
            self.z += self.vz

            if self.edges == "wrap":
                if self.x > (self._bounds['right'] + self.width / 2):
                    self.x = self._bounds['left'] - self.width / 2
                elif self.x < self._bounds['left'] - self.width / 2:
                    self.x = self._bounds['right'] + self.width / 2

                if(self.y > self._bounds['bottom'] + self.height / 2):
                    self.y = self._bounds['top'] - self.height / 2
                elif (self.y < self._bounds['top'] - self.height / 2):
                    self.y = self._bounds['bottom'] + self.height / 2

            elif self.edges == "bounce":
                if self.x > (self._bounds['right'] - self.width / 2):
                    self.x = self._bounds['right'] - self.width / 2
                    self.vx *= self.bounce
                elif self.x < (self._bounds['left'] + self.width / 2):
                    self.x = self._bounds['left'] + self.width / 2
                    self.vx *= self.bounce

                if self.y > (self._bounds['bottom'] - self.height / 2):
                    self.y = self._bounds['bottom'] - self.height / 2
                    self.vy *= self.bounce
                elif self.y < (self._bounds['top'] + self.height / 2):
                    self.y = self._bounds['top'] + self.height / 2
                    self.vy *= self.bounce

            # else if self.edges == "remove":
            #     if(self.x > self._bounds.right + self.width/2 || self.x<self._bounds.left - self.width/2 ||
            #        self.y > self._bounds.bottom + self.height/2 || self.y<self._bounds.top - self.height/2){

        if Global.isNaN(self.x):
            self.x = self.y = self.vx = self.vy = 0
